import sys,os
import utilities
from raptus.torii.python import Python

utilities = dict(sdir=utilities.sdir, ls=utilities.ls)
""" utilities are a set of helper functions. They will appear as globals
    in your python prompt. The globals can be extended with additional
    packages. To extend take a look in raptus.torii.plone.
"""
properties = dict()
""" properties are a set of helper attributes. Similar to the utilities, but
    properties are called by each connection. The call of the function performed
    in the context of the connection. This means you can use local attributes
    in your function, like app, arguments ... Only the return value is stored 
    in the globals. To extend take a look in raptus.torii.plone.
"""

scripts = dict(pack= '%s/scripts/pack.py' % os.path.dirname(__file__))
""" scripts can be run directly without the python prompt over torii. It's
    easy to customize you own scripts. Again please take a look
    at raptus.torii.plone
"""

interpreter = Python
""" The standard python interpreter. To create you own interpreter subclass
    interpreter.AbstractInterpreter and override all methods. Take a look at
    raptus.torii.python and raptus.torii.ipython.ipython
"""

tab_replacement = '    '




import sys
import code
import StringIO
from raptus.torii.interpreter import AbstractInterpreter

class InteractiveConsole(code.InteractiveConsole):
    errstream = StringIO.StringIO()
    def write(self, data):
        self.errstream.write(data)
    def flushErr(self):
        self.errstream = StringIO.StringIO()

class Outputcache(object):
    output = StringIO.StringIO()
    def __call__(self, arg):
        if arg is not None:
            self.output.write(arg)

    def flush(self):
        self.output = StringIO.StringIO()

class Python(AbstractInterpreter):
    """ default interpreter for torii
    """

    def __init__(self, locals):
        self.console = InteractiveConsole(locals)
        self.outputcache = Outputcache()

    def getPrompt1(self):
        return '>>> '

    def getPrompt2(self):
        return '... '

    def getPromptOut(self):
        return ''

    def getReadline(self):
        return None

    def resetStream(self):
        self.outputcache.flush()
        self.console.flushErr()
        
    def push(self, line):
        displayhook_old = sys.displayhook
        sys.displayhook = self.outputcache
        return self.console.push(line)
        sys.displayhook = displayhook_old
        
    def runcode(self,code):
        displayhook_old = sys.displayhook
        sys.displayhook = self.outputcache
        self.console.runcode(code)
        sys.displayhook = displayhook_old
    
    def complete(self, text):
        return []
    
    def getStdout(self):
        return self.outputcache.output
    
    def getErrorStream(self):
        return self.console.errstream
    
    def getSyntaxErrorStream(self):
        return StringIO.StringIO()
    





class AbstractInterpreter(object):
    
    def __init__(self, locals):
        pass

    def getPrompt1(self):
        """ return prompt like >>>
        """

    def getPrompt2(self):
        """ return prompt like ...
        """

    def getPromptOut(self):
        """ return prompt like Out[0]
        """

    def getReadline(self):
        """ return a  callable-object or a function to set the readline on the client
        """

    def resetStream(self):
        """ reset stdout and stderr stream. Called each time before
            a command is executed.
        """

    def push(self, line):
        """ push a single code line
        """
    
    def runcode(self,code):
        """ run a code-object
        """
    
    def complete(self, text):
        """ code completion
        """
    
    def getStdout(self):
        """ get the stdout for this cycle
        """
    
    def getErrorStream(self):
        """ return all generated errors in this cycle
        """
    def getSyntaxErrorStream(self):
        """ return all generated syntax-errors in this cycle
        """
    


import sys, os
import socket
from copy import copy
from ZServer.datatypes import ServerFactory
from ZServer import CONNECTION_LIMIT, requestCloseOnExec
import asyncore
from raptus.torii import config
from raptus.torii.conversation import Conversation

class ToriiServer(asyncore.dispatcher):
    def __init__(self, path,section, logger):
        asyncore.dispatcher.__init__(self)
        try:
            os.unlink(path)
        except os.error:
            pass
        self.threaded = section.threaded
        self.configuration = Configuration(section.extends, section.params)
        self.create_socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.bind(path)
        self.listen(1)
        self.socket.setblocking(True)
        self.log_info('ToriiServer is running\n\tSocketpath: %s' % path)
        
    def handle_accept(self):
        self.log_info('someone connected on torii server')
        connection, addr = self.accept()
        conversation = Conversation(connection, self.configuration)
        if self.threaded:
            conversation.start()
        else:
            conversation.run()
            
    def readable(self):
        return len(asyncore.socket_map) < CONNECTION_LIMIT

    def writable (self):
        return False
    
    def create_socket(self, family, type):
        asyncore.dispatcher.create_socket(self, family, type)


class Configuration(object):
    def __init__(self, extends, params):
        self.extends = self.parse_to_list(extends)
        self.params = self.parse_to_dict(params)
        self.utilities = copy(config.utilities)
        self.properties = copy(config.properties)
        self.scripts = copy(config.scripts)
        self.interpreter = config.interpreter
        for ext in self.extends:
            imp = __import__(ext, None, None, ext)
            
            if hasattr(imp,'initialize'):
                imp.initialize(self.params)
            if hasattr(imp,'utilities'):
                self.utilities.update(imp.utilities)
            if hasattr(imp,'properties'):
                self.properties.update(imp.properties)
            if hasattr(imp,'scripts'):
                self.scripts.update(imp.scripts)
            if hasattr(imp,'interpreter'):
                self.interpreter = imp.interpreter

    def parse_to_list(self, string):
        if not string:
            return []
        return [i.strip() for i in string.split(';')]

    def parse_to_dict(self, string):
        res = dict()
        if not string:
            return res
        for i in self.parse_to_list(string):
            k,v = i.split(':')
            res.update({k.strip():v.strip()})
        return res
    
    
class ToriiFactory(ServerFactory):
    
    def __init__(self, section):
        self.section = section
        
    def prepare(self, defaulthost='', dnsresolver=None, module=None, env=None, portbase=None):
        pass
        
    def create(self):
        from ZServer.AccessLogger import access_logger
        return ToriiServer(self.section.path.address, self.section, access_logger)





from App.config import getConfiguration
from raptus.torii import carrier

def getHumanSize(size):
        size
        if type(size) is type(''):
            return s

        if size >= 1048576.0:
            return '%.1fM' % (size/1048576.0)
        return '%.1fK' % (size/1024.0)


config = getConfiguration()

for db in config.databases:
    path = [i[1:] for i in db.getVirtualMountPaths()]
    mount = app
    if not '' in path:
        for i in path:
            mount = mount[i]
    db = mount._p_jar.db()
    size_before = getHumanSize(db.getSize())
    db.pack()
    size_after = getHumanSize(db.getSize())
    conversation(carrier.PrintText('Database: %s from %s to %s packed' % (db.database_name, size_before, size_after)))





import sys


class BaseCarrier(object):
    """ The communicatin between ZopeServer and Client goes over the instance BaseCarrier.
        The server send a pickle-instance to the client. The clinet recive the object and run
        a executable methode on this class. After all this the client send the object back
        to the ZopeServer.
    """
    
    def executable(self, client):
        """ this methode is called from the clinet
            each time while recived this object.
        """
        print "ooops you need to override this methode in the BaseCarrier"
        
class BaseCounterCarrier(object):
    """ CounterCarrier is a request from the client to the server.
    """
    def executable(self, interpreter):
        print 'override this methode in counterCarrier'
    
    
    
class FetchArguments(BaseCarrier):
    
    def executable(self, client):
        self.arguments = sys.argv

class BuildReadline(BaseCarrier):
    def __init__(self, readline):
        self.readline = readline
    
    def executable(self, client):
        if self.readline is not None:
            self.readline(client)

class GetCodeLine(BaseCarrier):
    
    def __init__(self, prompt ):
        self.prompt = prompt
    def executable(self, client):
        self.line = raw_input(self.prompt)
        

class GetNextCodeLine(GetCodeLine):
    
    def executable(self, client):
        self.line = raw_input(self.prompt)


class SendStdout(BaseCarrier):
    
    def __init__(self, stringIO, prompt=''):
        self.stringIO = stringIO
        self.prompt = prompt
    
    def executable(self, client):
        self.stringIO.seek(0)
        sys.stdout.write(self.prompt)
        for out in self.stringIO:
            print >> sys.stdout, out


class SendStderr(SendStdout):
    
    def executable(self, client):
        self.stringIO.seek(0)
        for out in self.stringIO:
            print >> sys.stderr, out

class PrintHelpText(BaseCarrier):
    
    def executable(self, client):
        print >> sys.stdout, client.__doc__

class PrintText(BaseCarrier):
    
    def __init__(self, text):
        self.text = text
    
    def executable(self, client):
        print >> sys.stdout, self.text

class ExitTorii(BaseCarrier):
    
    def executable(self, client):
        client.sock.close()
        sys.exit()

class FetchCompleter(BaseCounterCarrier):
    def __init__(self, text, state):
        self.text = text
        self.state = state
    def executable(self, interpreter):
        self.result = interpreter.complete(self.text)



import sys
import socket
import cPickle
import StringIO
from optparse import OptionParser
from raptus.torii import config

class Client(object):
    
    """
                                                            
            .....:::::::MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM   
  OMMMMMMMMMMMMMMMMMMMMMMM   MMMMMMMMMMMMMMMMMMMMMMM~MM     
   MM 8M MMM   M  MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM      
    MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM,NM MM                
             M MD                       M MM                
             M MM?                   ==MMIMMM88NMMM         
        MMMMMM MMMMMMMMMMMMMMMMMMMMMMMM MIMM   M  M         
        MMM MM MMMMMMMMMMMMMMMMMMMMMMMMMM MMMMMMMMM         
             M M                        M MM                
             MMM                        MIMM                
             MMM                        MIMM                
             M?M                        M MM                
             MMM                        M MM                
             MMM                        MIMM                
             MMM                        M~MM                
             MMM                        M7MM                
             7MM                        MIMM                
             7MM                        MIMM                
             IMM                        M7MM                
              MM                        MIMM                
             7MM                        MIMM                
             :::                        MDMM                
    
    
     Torii is a traditional Japanese gate most commonly 
     found at the entrance of or within a Shinto shrine.
     But in the zopy-world raptus.torii provide a gate to a
     running zope-server (holy-zopy). The access to the server can be
     get of differnt ways. here the manual:


     help          this help
                  
     debug         interactive mode
                 
     list          summary-list of all available scripts

     run <script>  run the given script



    """
              
    
    def __init__(self, path):
        self.path = path
        
    
    def main(self):
        try:
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.sock.connect(self.path)
            while True:
                carrier = cPickle.load(self.sock.makefile())
                carrier.executable(self)
                cPickle.dump(carrier, self.sock.makefile())
        except socket.error, msg:
                    print >> sys.stdout, msg
        except KeyboardInterrupt:
            pass
        print >> sys.stdout, '\n'
        self.sock.close()





# ohhh we have a loop around the imports
# workaround: only import torii instead the module carrier
from raptus import torii

def sdir(value):
    biggest = 0
    att = dir(value)
    for i in att:
        if biggest < len(i):
            biggest = len(i)
    for i in att:
        try:
            content = repr(getattr(value,i))[:60]
        except:
            content = '????'
            pass
        conversation(torii.carrier.PrintText(i+' '* (biggest - len(i))+content))


def ls(node):
    for i in node.getChildNodes():
        conversation(torii.carrier.PrintText(repr(i)))



import sys, os
import cPickle
import threading
import Zope2
import Globals
from copy import copy
from Testing.makerequest import makerequest
from codeop import compile_command
from raptus.torii import config
from raptus.torii import carrier


class Conversation(threading.Thread):
    
    def __init__(self, connection, configuration):
        super(Conversation, self).__init__()
        self.connection = connection
        self.configuration = configuration
        self.locals = copy(configuration.utilities)
        
        self.interpreter = configuration.interpreter(self.locals)

        self.arguments = self.conversation(carrier.FetchArguments()).arguments
        self.locals.update(arguments=self.arguments)
        self.locals.update(conversation=self.conversation)
        self.conversation(carrier.BuildReadline(self.interpreter.getReadline()))
    def run(self):
        dbConnection = Zope2.DB.open()
        application=Zope2.bobo_application(connection=dbConnection)
        application=makerequest(application)
        self.locals.update(app=application)

        for func in self.configuration.utilities.values():
            getattr(func, 'func_globals', {}).update(self.locals)

        for name, func in self.configuration.properties.items():
            if hasattr(func, 'func_code'):
                di = func.func_globals
                di.update(self.locals)
                res = eval(func.func_code,di)
                self.locals.update({name:res})
                continue
            try:
                res = func()
                self.locals.update({name:res})
                continue
            except:
                self.locals.update({name:func})

        mode = dict(    help = lambda: self.conversation(carrier.PrintHelpText()),
                        debug = self.interactiveMode,
                        list = self.listScripts,
                        run = self.runScript,
                   )
        if not Globals.DevelopmentMode:
            msg  = """
                       Sorry, but run the debug mode on productive Zope is too risky!
                       Pleas stop your Server and run it in the Foreground (fg) mode.
                   """
            mode.update(dict(debug=lambda: self.conversation(carrier.PrintText(msg))))


        try:
            if len(self.arguments) > 1 and self.arguments[1] in mode:
                mode[self.arguments[1]]()
            else:
                mode['help']()
            self.conversation(carrier.ExitTorii())
                
        except:
            dbConnection.transaction_manager.abort()
            dbConnection.close()
            self.connection.close()

    def listScripts(self):
        self.conversation(carrier.PrintText('here is a list with all available scripts:\n'))
        for name, path in self.configuration.scripts.items():
            self.conversation(carrier.PrintText(' '*10+name))

    def runScript(self):
        name = self.arguments[2]
        path = self.configuration.scripts.get(name, None)
        if path is None:
            self.conversation(carrier.PrintText("Sorry, but this script dosen't exist"))
            return
        f = file(path)
        code = compile_command(f.read(),path,'exec')
        if code is None:
            self.conversation(carrier.PrintText('script is not finished, blank line at the end missing?'))
        f.close()
        self.interpreter.runcode(code)
        stderr = self.interpreter.getErrorStream()
        if stderr.len:
            self.conversation(carrier.SendStderr(stderr))

    def interactiveMode(self):
        self.conversation(carrier.PrintText('Available global variables:'))
        for key in self.locals.keys():
            self.conversation(carrier.PrintText(key))
        
        while True:
            self.interpreter.resetStream()
            
            input = self.conversation(carrier.GetCodeLine(self.interpreter.getPrompt1()))
            try:
                while self.interpreter.push(input.line):
                    input = self.conversation(carrier.GetNextCodeLine(self.interpreter.getPrompt2()))
            except Exception, mesg:
                pass

            stderr = self.interpreter.getErrorStream()
            if stderr.len:
                self.conversation(carrier.SendStderr(stderr))
            stderr = self.interpreter.getSyntaxErrorStream()
            if stderr.len:
                self.conversation(carrier.SendStderr(stderr))
            stdout = self.interpreter.getStdout()
            if stdout.len:
                self.conversation(carrier.SendStdout(stdout, self.interpreter.getPromptOut()))


    def conversation(self, carrierObject):
        cPickle.dump(carrierObject, self.connection.makefile())
        while True:
            obj = cPickle.load(self.connection.makefile())
            if isinstance(obj, carrier.BaseCarrier):
                break;
            else:
                obj.executable(self.interpreter)
                cPickle.dump(obj, self.connection.makefile())
        return obj

    



from setuptools import setup, find_packages
import os

version = '1.0b2'

setup(name='raptus.torii',
      version=version,
      description='Torii allows access to a running zope instance over a python prompt',
      long_description=open("README.txt").read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Framework :: Plone",
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Framework :: ZODB",
        "Programming Language :: Python",
        ],
      keywords='zope remote prompt zodb',
      author='sriolo',
      author_email='sriolo@raptus.com',
      url='http://svn.plone.org/svn/collective/raptus.torii',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['raptus'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      setup_requires=[ ],
      paster_plugins=[ ],
      )
